#import <Foundation/Foundation.h>

#import <NisusCore/NisusDefines.h>
#import <NisusCore/NSObject+NisusCore.h>
#include <tgmath.h>




/********************
 *** Misc Defines ***
 *******************/

#define NISUS_UNUSED(var) (void)(var)

#define NISUS_SWAP_SCALARs(xxx,yyy)				\
	{											\
		__typeof__(xxx) xxxTemp = (xxx);		\
		xxx = yyy;								\
		yyy = xxxTemp;							\
	}

// Returns "primaryExpr" if its value is *not* nil, otherwise returns "fallbackExpr".
// This behavior is short-circuited, so that "fallbackExpr" is not evaluated unless the primary value is nil.
#define NISUS_NIL_FALLBACK( primaryExpr, fallbackExpr )				\
	({																\
		__typeof__(primaryExpr) nisusValue = (primaryExpr);			\
		if( nil == nisusValue ) {									\
			nisusValue = (fallbackExpr);							\
		}															\
		nisusValue;													\
	})

// Checks if the given object's kind matches the given class. If so, returns the object, otherwise returns nil.
@interface NSObject (NisusCoreCast)
+ (instancetype) conditionalCastNisus:(id)obj ;
@end

#define NISUS_CAST_TO_CLASS( classType, objectExpr )				\
	({																\
		classType* nisusValue = (classType*)(objectExpr);			\
		if( ! [nisusValue isKindOfClass:[classType class]] ) {		\
			nisusValue = nil;										\
		}															\
		nisusValue;													\
	})

static NISUS_INLINE void NisusAccessIfResponds( id testObject, SEL selector, NS_NOESCAPE void(^accessor)(id object) )
{
	if( [testObject respondsToSelector:selector] ) {
		accessor( testObject );
	}
}

#define NISUS_NOT_NULL_INVOKE(block, ...)		(NULL != block) ? block(__VA_ARGS__) : nil

#define NISUS_NOT_NULL_LVALUE(pointer)			if( NULL != (pointer) ) *(pointer)
#define NISUS_NOT_NULL_SET(pointer,value)	  { if( NULL != (pointer) ) *(pointer) = (value); }
#define NisusSetNotNullPointer(pointer,value) { if( NULL != (pointer) ) *(pointer) = (value); }

#define NISUS_NOT_NULL_FREE(ptr)			  { if(NULL != (ptr)) free(ptr); }
#define NISUS_NOT_NULL_FREE_CLEAR(ptr)	\
	if( NULL != ptr ) {					\
		free(ptr);						\
		ptr = NULL;						\
	}

// Ensures that both (min <= value) and (value <= max)
#define NISUS_CLAMP( valueExpr, minExpr, maxExpr )					\
	({																\
		__typeof__(valueExpr) clampValue = (valueExpr);				\
		__typeof__(valueExpr) clampMin = (minExpr);					\
		__typeof__(valueExpr) clampMax = (maxExpr);					\
																	\
		__typeof__(valueExpr) clampResult = clampValue;				\
		if( clampResult < clampMin ) clampResult = clampMin;		\
		if( clampResult > clampMax ) clampResult = clampMax;		\
																	\
		clampResult;												\
	})

// This is a CLAMP for enums that define a "_MIN" and "_MAX" macro. It also handles type-casting.
#define NISUS_CLAMP_ENUM( valueExpr, enumTypeName )					\
	(enumTypeName)(NISUS_CLAMP( valueExpr, enumTypeName ## _MIN, enumTypeName ## _MAX ))

// This is a CLAMP for enums that define a "_MIN" and "_MAX" macro. It also handles type-casting.
#if NISUS_DEBUG
	#define NISUS_CAST_ENUM( enumTypeName, valueExpr )				\
	({																\
		enumTypeName clampValue = (valueExpr);						\
		enumTypeName clampMin = (enumTypeName ## _MIN);				\
		enumTypeName clampMax = (enumTypeName ## _MAX);				\
																	\
		NISUS_ASSERT_SOFT( (valueExpr) <= (enumTypeName ## _MAX) );	\
		NISUS_ASSERT_SOFT( (valueExpr) >= (enumTypeName ## _MIN) );	\
																	\
		enumTypeName clampResult = clampValue;						\
		if( clampResult < clampMin ) clampResult = clampMin;		\
		if( clampResult > clampMax ) clampResult = clampMax;		\
																	\
		clampResult;												\
	})
#else
	#define NISUS_CAST_ENUM( enumTypeName, valueExpr )					\
		(enumTypeName)(NISUS_CLAMP( valueExpr, enumTypeName ## _MIN, enumTypeName ## _MAX ))
#endif

// This defers calling the given block until the NSRunLoop unwinds, so no code is active when the block is triggered (aside from basic RunLoop machinery).
// This is like performSelector with a delay of 0, but that executes immediately on 10.14 instead of being delayed.
extern void NisusDeferToBaseRunLoop( void(^deferBlock)(void) );



/******************
 *** Bit Flags ***
 *****************/

// tests if any of the given test flags are set, which is the same as: tests if a single bit flag is set.
#define NisusIsBitFlagSet( flags, testFlag )			(0 != ((flags) & (testFlag)))

// tests if all flags are set
#define NisusIsAllBitFlagsSet( flags, testFlags )		((testFlags) == ((flags) & (testFlags)))

// tests if only bits within "testFlags" are set
#define NisusIsOnlyBitFlagsSet( flags, testFlags )		(0 == (flags & (~testFlags)))

// adding/removing flags
#define NisusAddBitFlags( flags, addFlags )			((flags) | (addFlags))
#define NisusRemoveBitFlags( flags, remFlags )		((flags) & (~(remFlags)))

#define NisusAddBitFlagsIf( isModify, flags, addFlags )			((isModify) ? NisusAddBitFlags(flags,addFlags) : (flags))
#define NisusRemoveBitFlagsIf( isModify, flags, remFlags )		((isModify) ? NisusRemoveBitFlags(flags,addFlags) : (flags))
#define NisusToggleBitFlag( flags, toggleFlag )					((flags) ^ (toggleFlag))

// Returns the indexes (between 0-63) of any bits set in the given flags, eg: returns (2) if the flags are 0b0100, or (0,3) if the flags are 0b1001
NSIndexSet* NisusIndexesForBitFlags( NSUInteger flags );




/******************
 **** Equality ****
 *****************/

#define NisusIsEqualFloats( f1, f2, tolerance ) ( fabs((f1) - (f2)) < (tolerance) )

// Just like sending isEqual, but handles annoying nil cases (as well as shortcut for pointer equality).
static NISUS_INLINE BOOL NisusIsEqualObjects( id obj1, id obj2 )
{
	if( obj1 == obj2 ) return YES;
	if( nil == obj1 ) return [obj2 isEqual:obj1];
	return [obj1 isEqual:obj2];
}

// Helper for implementing -isEqual: method which handles some early bailout conditions.
// Checks to make sure the given object is the desired class. If not then returns NO early.
#define NisusIsEqual_IMPL_ENSURE_CLASS( otherObjectExpr, expectedClass )	\
{																			\
	id otherObject = (otherObjectExpr);										\
	if( nil == otherObject ) return NO;										\
	if( self == otherObject ) return YES;									\
	if( ! [otherObject isKindOfClass:expectedClass] ) return NO;			\
}

// Ordering any kind of directly-comparable C scalar values, eg: ints, pointers, etc
#define NisusCompareScalars( xxxExpr, yyyExpr )										\
		({																			\
			__typeof__(xxxExpr) xxxValue = (xxxExpr);								\
			__typeof__(yyyExpr) yyyValue = (yyyExpr);								\
			\
			NSComparisonResult nisusCompRet = NSOrderedSame;						\
			if( xxxValue < yyyValue ) nisusCompRet = NSOrderedAscending;			\
			else if( xxxValue > yyyValue ) nisusCompRet = NSOrderedDescending;		\
			\
			nisusCompRet;															\
		})

// Handles ordering pointers.
static NISUS_INLINE NSComparisonResult NisusComparePointers( void* p1, void* p2 )
{
	NSComparisonResult result = NisusCompareScalars(p1, p2);
	return result;
}

